// ***************************************************************************
// Implementation System Includes
// ***************************************************************************

#include <be/interface/Bitmap.h>
#include <be/interface/Window.h>
#include <be/media/MediaNode.h>

#include <posix/stdio.h>
#include <posix/string.h>

// ***************************************************************************
// Implementation Project Includes
// ***************************************************************************

#include "EavesdropperMain.h"
#include "IconFactory.h"
#include "InfoWindow.h"
#include "AbstractNodeView.h"

// ***************************************************************************
// Implementation Constant Definitions
// ***************************************************************************

// ***************************************************************************
// Implementation Type Definitions
// ***************************************************************************

// ***************************************************************************
// Implementation Variable Definitions
// ***************************************************************************

// ***************************************************************************
// Implementation Function Declarations
// ***************************************************************************

// ***************************************************************************
// Global Variable Definitions
// ***************************************************************************

// ***************************************************************************
// PUBLIC Member Function Definitions
// ***************************************************************************

// ===========================================================================
// Constructor
// ===========================================================================

AbstractNodeView::AbstractNodeView(
		InfoWindow* infoWindow,
		const char* nodeName)
	: BView(BRect(0, 0, 1, 1),
		"AbstractNodeView",
		B_FOLLOW_NONE,
		B_WILL_DRAW) 
{
	// Initialize attributes.
	this->infoWindow = infoWindow;
	isSelected = false;
	isRollover = false;

	SetViewColor(COLOR_BACKGROUND);

	// Cache info about the view's font.
	GetFontHeight(&fontHeight);
	fontSize = fontHeight.ascent + fontHeight.descent + fontHeight.leading;

	// Split up the node's name into chunks.
	nameChunks = new BList(3);
	SplitUpName(nodeName);
	
	// Figure out what the dimensions of the view will be.
	
	float width = 1				// Selected/rollover border
		+ 1						// Normal border
		+ 2						// Spacing
		+ 16 + 2 + 16 + 2 + 16	// The node's kinds
		+ 2						// Spacing
		+ 1						// Selected/rollover border
		+ 1;					// Normal border

	float height = 1 + 1		// Selected/rollover border + normal border
		+ fontSize				// The node's id
		+ 1						// Border between node id and kinds
		+ 2						// Spacing
		+ 16 + 2 + 16 + 2 + 16	// The node's kinds
		+ 2						// Spacing
		+ 1 + 1					// Selected/rollover border + normal border
		+ (nameChunks->CountItems() * fontSize); // The node's name

	// And stretch the view to its new size.
	ResizeTo(width, height);
}

// ===========================================================================
// Destructor
// ===========================================================================

AbstractNodeView::~AbstractNodeView(void)
{
	// Delete the list of name chunks.
	for (int32 t=0; t<nameChunks->CountItems(); ++t)
	{
		free(nameChunks->ItemAt(t));
	}
	delete nameChunks;
}

// ===========================================================================
// MouseDown
// ===========================================================================

void
AbstractNodeView::MouseDown(BPoint where)
{
	// Ask the live nodes view to select us.
	if (!isSelected)
	{
		infoWindow->SelectNode(this);
	}
}

// ===========================================================================
// MouseMoved
// ===========================================================================

void
AbstractNodeView::MouseMoved(BPoint point,
	uint32 transit,
	const BMessage* message)
{
	// Don't put the node in the rollover state if the user is dragging
	// another node on top of us.
	if (message != 0)
	{
		return;
	}

	// Figure out whether the mouse is currently rolling over the view.
	if (transit == B_ENTERED_VIEW)
	{
		isRollover = true;

		if (!isSelected)
		{
			Invalidate();
		}
	}
	else if (transit == B_EXITED_VIEW)
	{
		isRollover = false;

		if (!isSelected)
		{
			Invalidate();
		}
	}
}

// ===========================================================================
// SetSelected
// ===========================================================================

void
AbstractNodeView::SetSelected(bool flag)
{
	isSelected = flag;
	Invalidate();
}

// ===========================================================================
// IsSelected
// ===========================================================================

bool
AbstractNodeView::IsSelected()
{
	return isSelected;
}

// ***************************************************************************
// PROTECTED Member Function Definitions
// ***************************************************************************

// ===========================================================================
// DrawNode
// ===========================================================================

void 
AbstractNodeView::DrawNode(int32 nodeId, int32 nodeKinds)
{
	SetDrawingMode(B_OP_OVER);
	
	char string[1024];

	BRect bounds = Bounds();
	float width  = bounds.Width();
	float height = bounds.Height();

	float borderTop    = 0;
	float borderLeft   = (width - 60)/2;
	float borderRight  = borderLeft + 60;
	float borderBottom = fontSize + 61;

	// Draw the background for the node's id.
	SetHighColor(COLOR_TAB);
	FillRect(BRect(borderLeft + 1,
		borderTop + 1,
		borderRight - 1,
		fontSize + 2));

	// Draw the background for the node's kinds.
	SetHighColor(COLOR_NODE);
	FillRect(BRect(borderLeft + 1,
		fontSize + 4,
		borderRight - 1,
		borderBottom - 1));

	// Draw a border around the node's id and kinds.
	SetHighColor(COLOR_BORDER);
	StrokeRect(BRect(borderLeft + 1,
		borderTop + 1,
		borderRight - 1,
		borderBottom - 1));

	StrokeLine(
		BPoint(borderLeft + 1, fontSize + 3),
		BPoint(borderRight - 1, fontSize + 3));

	// Draw an extra border around the node if it is selected or rollover'ed.
	if (isSelected)
	{
		SetHighColor(COLOR_SELECTED);
		StrokeRect(BRect(borderLeft,
			borderTop,
			borderRight,
			borderBottom));
	}
	else if (isRollover)
	{
		SetHighColor(COLOR_ROLLOVER);
		StrokeRect(BRect(borderLeft,
			borderTop,
			borderRight,
			borderBottom));
	}
	
	// Calculate the position of the id.
	sprintf(string, "%ld", nodeId);
	float idX = (width - StringWidth(string))/2;
	float idY = fontHeight.ascent + 2;

	SetHighColor(COLOR_TEXT);
	DrawString(string, BPoint(idX, idY));

	// Display this icons for the node's kinds.
	if (nodeKinds & B_BUFFER_PRODUCER)
	{
		DrawBitmap(IconFactory::GetBitmapFor(IconFactory::BUFFER_PRODUCER),
			BPoint(borderLeft + 4, fontSize + 5));
	}        
	if (nodeKinds & B_BUFFER_CONSUMER)
	{
		DrawBitmap(IconFactory::GetBitmapFor(IconFactory::BUFFER_CONSUMER),
			BPoint(borderLeft + 22, fontSize + 5));
	}        
	if (nodeKinds & B_TIME_SOURCE)
	{
		DrawBitmap(IconFactory::GetBitmapFor(IconFactory::TIME_SOURCE),
			BPoint(borderLeft + 40, fontSize + 5));
	}        
	if (nodeKinds & B_CONTROLLABLE)
	{
		DrawBitmap(IconFactory::GetBitmapFor(IconFactory::CONTROLLABLE),
			BPoint(borderLeft + 4, fontSize + 23));
	}        
	if (nodeKinds & B_FILE_INTERFACE)
	{
		DrawBitmap(IconFactory::GetBitmapFor(IconFactory::FILE_INTERFACE),
			BPoint(borderLeft + 22, fontSize + 23));
	}        
	if (nodeKinds & B_ENTITY_INTERFACE)
	{
		// Not implemented yet.
	}        
	if (nodeKinds & B_PHYSICAL_INPUT)
	{
		DrawBitmap(IconFactory::GetBitmapFor(IconFactory::PHYSICAL_INPUT),
			BPoint(borderLeft + 4, fontSize + 41));
	}        
	if (nodeKinds & B_PHYSICAL_OUTPUT)
	{
		DrawBitmap(IconFactory::GetBitmapFor(IconFactory::PHYSICAL_OUTPUT),
			BPoint(borderLeft + 22, fontSize + 41));
	}        
	if (nodeKinds & B_SYSTEM_MIXER)
	{
		DrawBitmap(IconFactory::GetBitmapFor(IconFactory::SYSTEM_MIXER),
			BPoint(borderLeft + 40, fontSize + 41));
	}

	// Draw the node's name.
	SetHighColor(COLOR_TEXT);
	for (int32 t=0; t<nameChunks->CountItems(); ++t)
	{
		const char* nameChunk = (const char*) nameChunks->ItemAt(t);
		float nameX = (width - StringWidth(nameChunk))/2;
		float nameY = fontSize + 61 + fontHeight.ascent + t*fontSize;

		DrawString(nameChunk, BPoint(nameX, nameY));
    }	
}

// ===========================================================================
// SplitUpName
// ===========================================================================

void
AbstractNodeView::SplitUpName(const char* nodeName)
{
	// Maximum length of a name chunk in pixels.
	const float MAX_CHUNK_LENGTH = 60;
	
	// Get a copy of the source string so we can modify it at will.
	char* name = strdup(nodeName);
	char* start = name;
	char* end = start;
	char* ptr = start;
		
	while (1)
	{
		// Read the next character from the source string.
		char c = *ptr;
		
		// Not a space and not end of string? Then continue looking.
		if ((c != ' ') && (c != '\000'))
		{
			++ptr;
		}
		else
		{
			// Mark the end of the chunk. We need to do that here,
			// or StringWidth() will go berserk.
			*ptr = '\000';

			// Calculate the width of the string in pixels.
			float stringWidth = StringWidth(start);

			// Does the chunk no longer fit? Or have we reached the
			// end of the string?
			if ((stringWidth > MAX_CHUNK_LENGTH) || (c == '\000'))
			{
				if ((end == start) || (stringWidth <= MAX_CHUNK_LENGTH))
				{
					AddNameChunk(strdup(start));
					
					if (c == '\000')
					{
						break;
					}
					
					++ptr;
					start = ptr;
					end = ptr;
				}
				else
				{
					*end = '\000';
					AddNameChunk(strdup(start));
					
					++end;
					start = end;
					*ptr = c;
				}
			}
			
			// There is still place in this chunk, so add the space to it.
			// (This overwrites the null character we added earlier.)
			else
			{
				end = ptr;
				*ptr = c;
				++ptr;
			}
		}
	}
}

// ===========================================================================
// AddNameChunk
// ===========================================================================

void
AbstractNodeView::AddNameChunk(const char* chunk)
{
	char* truncatedChunk = (char*) malloc(strlen(chunk)*3);

	BFont font;
	GetFont(&font);
	font.GetTruncatedStrings(&chunk,
		1,
		B_TRUNCATE_MIDDLE,
		60,
		&truncatedChunk);

	nameChunks->AddItem(truncatedChunk);
}

// ***************************************************************************
// PRIVATE Member Function Definitions
// ***************************************************************************

// ***************************************************************************
// Helper Function Definitions
// ***************************************************************************

// ***************************************************************************
// Global Function Definitions
// ***************************************************************************

// ***************************************************************************
// Implementation Function Definitions
// ***************************************************************************
